"""
Unit tests for interface
"""
# System imports
import os
import time
import unittest
import jsondiff
import sys

# airr imports
import airr
from airr.schema import ValidationError

# Paths
test_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(test_path, 'data')


class TestInferface(unittest.TestCase):
    def setUp(self):
        print('-------> %s()' % self.id())

        # Test data
        self.rearrangement_good = os.path.join(data_path, 'good_rearrangement.tsv')
        self.rearrangement_bad = os.path.join(data_path, 'bad_rearrangement.tsv')
        self.rep_good = os.path.join(data_path, 'good_repertoire.yaml')
        self.rep_bad = os.path.join(data_path, 'bad_repertoire.yaml')
        self.germline_good = os.path.join(data_path, 'good_germline_set.json')
        self.germline_bad = os.path.join(data_path, 'bad_germline_set.json')
        self.genotype_good = os.path.join(data_path, 'good_genotype_set.json')
        self.genotype_bad = os.path.join(data_path, 'bad_genotype_set.json')
        self.combined_yaml = os.path.join(data_path, 'good_combined_airr.yaml')
        self.combined_json = os.path.join(data_path, 'good_combined_airr.json')

        # Output data
        self.output_rep = os.path.join(data_path, 'output_rep.json')
        self.output_good = os.path.join(data_path, 'output_data.json')
        self.output_blank = os.path.join(data_path, 'output_blank.json')

        # Expected output
        self.shape_good = (9, 44)
        self.shape_bad = (9, 44)

        # Start timer
        self.start = time.time()

    def tearDown(self):
        t = time.time() - self.start
        print('<- %.3f %s()' % (t, self.id()))

    # @unittest.skip('-> load(): skipped\n')
    def test_load_rearrangement(self):
        # Good data
        result = airr.load_rearrangement(self.rearrangement_good)
        self.assertTupleEqual(result.shape, self.shape_good, 'load(): good data failed')

        # Bad data
        result = airr.load_rearrangement(self.rearrangement_bad)
        self.assertTupleEqual(result.shape, self.shape_bad, 'load(): bad data failed')

    # @unittest.skip('-> repertoire_template(): skipped\n')
    def test_repertoire_template(self):
        try:
            with self.assertWarns(DeprecationWarning, msg='repertoire_template(): failed to issue DeprecationWarning'):
                rep = airr.repertoire_template()
            airr.write_airr(self.output_blank, {'Repertoire': rep}, validate=False, debug=True)
        except:
            pass

    # @unittest.skip('-> schema.template(): skipped\n')
    def test_schema_template(self):
        # Repertoire template
        try:
            data = airr.schema.RepertoireSchema.template()
            valid = airr.schema.RepertoireSchema.validate_object(data)
            self.assertTrue(valid, 'Schema.template("Repertoire"): repertoire template failed validation')
        except:
            self.assertTrue(False, 'Schema.template("Repertoire"): repertoire template failed validation')

        # GermlineSet template
        try:
            data = airr.schema.GermlineSetSchema.template()
            valid = airr.schema.GermlineSetSchema.validate_object(data)
            self.assertTrue(valid, 'Schema.template("GermlineSet"): repertoire template failed validation')
        except:
            self.assertTrue(False, 'Schema.template("GermlineSet"): repertoire template failed validation')

         # GenotypeSet template
        try:
            data = airr.schema.GenotypeSetSchema.template()
            valid = airr.schema.GenotypeSetSchema.validate_object(data)
            self.assertTrue(valid, 'Schema.template("GenotypeSet"): repertoire template failed validation')
        except:
            self.assertTrue(False, 'Schema.template("GenotypeSet"): repertoire template failed validation')

    # @unittest.skip('-> validate(): skipped\n')
    def test_validate_rearrangement(self):
        # Good data
        try:
            result = airr.validate_rearrangement(self.rearrangement_good)
            self.assertTrue(result, 'validate(): good data failed')
        except:
            self.assertTrue(False, 'validate(): good data failed')

        # Bad data
        try:
            result = airr.validate_rearrangement(self.rearrangement_bad)
            self.assertFalse(result, 'validate(): bad data failed')
        except Exception as inst:
            print(type(inst))
            raise inst

    # @unittest.skip('-> read_airr(): skipped\n')
    def test_read_airr(self):
        # Good data
        print('--> Good data')
        try:
            data = airr.read_airr(self.rep_good, validate=True, debug=True)
        except:
            self.fail('read_airr(): good data failed')

        # Bad data
        print('--> Bad data')
        with self.assertRaises(ValidationError, msg="read_airr(): bad data passed validation"):
            data = airr.read_airr(self.rep_bad, validate=True, debug=True)

        # Combined yaml
        print('--> Combined YAML')
        try:
            data_yaml = airr.read_airr(self.combined_yaml, validate=True, debug=True)
        except:
            self.fail('read_airr(): combined yaml failed')

        # Combined json
        print('--> Combined JSON')
        try:
            data_json = airr.read_airr(self.combined_json, validate=True, debug=True)
        except:
            self.fail('read_airr(): combined json failed')

        # Check equality of yaml and json
        self.assertDictEqual(data_yaml, data_json, msg="read_airr(): yaml and json imports are not equal")


    # @unittest.skip('-> validate_airr(): skipped\n')
    def test_validate_airr(self):
        # Good data
        print('--> Good data')
        # As array
        try:
            data = airr.read_airr(self.rep_good, validate=True, debug=True)
            valid = airr.validate_airr(data, debug=True)
            self.assertTrue(valid, 'validate_airr(): good data array failed')
        except:
            self.assertTrue(False, 'validate_airr(): good data array failed')

        # As dict
        try:
            array = airr.read_airr(self.rep_good, validate=False, debug=False)
            data = {'Repertoire': {x['repertoire_id']: x for x in array['Repertoire']}}
            valid = airr.validate_airr(data, debug=True)
            self.assertTrue(valid, 'validate_airr(): good data dict failed')
        except:
            self.assertTrue(False, 'validate_airr(): good data dict failed')

        # Bad data
        print('--> Bad data')
        # As array
        try:
            data = airr.read_airr(self.rep_bad, validate=True, debug=True)
            valid = airr.validate_airr(data, debug=True)
            self.assertFalse(valid, 'validate_airr(): bad data array failed')
        except ValidationError:
            pass
        except Exception as inst:
            print(type(inst))
            raise inst

        # As dict
        try:
            array = airr.read_airr(self.rep_bad, validate=False, debug=False)
            data = {'Repertoire': {x['repertoire_id']: x for x in array['Repertoire']}}
            valid = airr.validate_airr(data, debug=True)
            self.assertFalse(valid, 'validate_airr(): bad data dict failed')
        except ValidationError:
            pass
        except Exception as inst:
            print(type(inst))
            raise inst

    # @unittest.skip('-> load_repertoire(): skipped\n')
    def test_load_repertoire(self):
        # Good data
        try:
            with self.assertWarns(DeprecationWarning, msg='load_repertoire(): failed to issue DeprecationWarning'):
                data = airr.load_repertoire(self.rep_good, validate=True, debug=True)
        except:
            self.assertTrue(False, 'load_repertoire(): good data failed')

        # Bad data
        try:
            with self.assertWarns(DeprecationWarning, msg='load_repertoire(): failed to issue DeprecationWarning'):
                data = airr.load_repertoire(self.rep_bad, validate=True, debug=True)
            self.assertFalse(True, 'load_repertoire(): bad data passed')
        except ValidationError:
            pass
        except Exception as inst:
            print(type(inst))
            raise inst

    # @unittest.skip('-> write_repertoire(): skipped\n')
    def test_write_repertoire(self):
        # Good data
        try:
            with self.assertWarns(DeprecationWarning, msg='load_repertoire(): failed to issue DeprecationWarning'):
                data = airr.load_repertoire(self.rep_good, validate=True, debug=True)
            with self.assertWarns(DeprecationWarning, msg='write_repertoire(): failed to issue DeprecationWarning'):
                result = airr.write_repertoire(self.output_rep, data['Repertoire'], debug=True)
            with self.assertWarns(DeprecationWarning, msg='load_repertoire(): failed to issue DeprecationWarning'):
                # verify we can read it
                obj = airr.load_repertoire(self.output_rep, validate=True, debug=True)

            # is the data identical?
            if jsondiff.diff(obj['Repertoire'], data['Repertoire']) != {}:
                print('Output data does not match', file=sys.stderr)
                print(jsondiff.diff(obj, data), file=sys.stderr)
                self.assertTrue(False, 'write_repertoire(): Output data does not match')
        except:
            self.assertTrue(False, 'write_repertoire(): good data failed')

    # @unittest.skip('-> load_germline(): skipped\n')
    def test_read_germline(self):
        # Good data
        try:
            result = airr.read_airr(self.germline_good, validate=True, debug=True)
        except ValidationError:
            self.assertTrue(False, 'load_germline(): good data failed')

        # Bad data
        try:
            result = airr.read_airr(self.germline_bad, validate=True, debug=True)
            self.assertFalse(True, 'load_germline(): bad data succeeded')
        except ValidationError:
            pass

    # @unittest.skip('-> validate_germline(): skipped\n')
    def test_validate_germline(self):
        # Good data
        print('--> Good data')
        try:
            result = airr.read_airr(self.germline_good, validate=True, debug=True)
            valid = airr.validate_airr(result, debug=True)
            self.assertTrue(valid, 'validate_germline(): good data failed')
        except ValidationError:
            self.assertTrue(False, 'validate_germline(): good data failed')

        # Bad data
        print('--> Bad data')
        try:
            result = airr.read_airr(self.germline_bad, validate=True, debug=True)
            valid = airr.validate_airr(result, debug=True)
            self.assertFalse(valid, 'validate_germline(): bad data succeeded')
        except ValidationError:
            pass

    # @unittest.skip('-> load_genotype(): skipped\n')
    def test_read_genotype(self):
        # Good data
        print('--> Good data')
        try:
            result = airr.read_airr(self.genotype_good, validate=True, debug=True)
        except ValidationError:
            self.assertTrue(False, 'load_genotype(): good data failed')

        # Bad data
        print('--> Bad data')
        try:
            result = airr.read_airr(self.genotype_bad, validate=True, debug=True)
            self.assertFalse(True, 'load_genotype(): bad data succeeded')
        except ValidationError:
            pass

    # @unittest.skip('-> validate_genotype(): skipped\n')
    def test_validate_genotype(self):
        # Good data
        print('--> Good data')
        try:
            result = airr.read_airr(self.genotype_good, validate=True, debug=True)
            valid = airr.validate_airr(result, debug=True)
            self.assertTrue(valid, 'validate_genotype(): good data failed')
        except ValidationError:
            self.assertTrue(False, 'validate_genotype(): good data failed')

        # Bad data
        print('--> Bad data')
        try:
            result = airr.read_airr(self.genotype_bad, validate=True, debug=True)
            valid = airr.validate_airr(result, debug=True)
            self.assertFalse(valid, 'validate_genotype(): bad data succeeded')
        except ValidationError:
            pass

    # @unittest.skip('-> load_genotype(): skipped\n')
    def test_write_airr(self):
        # Good data as array
        try:
            repertoire_data = airr.read_airr(self.rep_good, validate=True, debug=True)
            germline_data = airr.read_airr(self.germline_good, validate=True, debug=True)
            genotype_data = airr.read_airr(self.genotype_good, validate=True, debug=True)

            # combine together and write
            obj = {}
            obj['Repertoire'] = repertoire_data['Repertoire']
            obj['GermlineSet'] = germline_data['GermlineSet']
            obj['GenotypeSet'] = genotype_data['GenotypeSet']
            airr.write_airr(self.output_good, obj, validate=True, debug=True)

            # verify we can read it
            data = airr.read_airr(self.output_good, validate=True, debug=True)

            # is the data identical?
            del data['Info']
            if jsondiff.diff(obj, data) != {}:
                print('Output data does not match', file=sys.stderr)
                print(jsondiff.diff(obj, data), file=sys.stderr)
                self.assertTrue(False, 'write_airr_data(): Output data does not match')

        except Exception as inst:
            self.assertTrue(False, 'write_airr_data(): good data failed')
            print(type(inst))
            raise inst

        # Good data as dict
        try:
            # Load data
            repertoire_array = airr.read_airr(self.rep_good, validate=True, debug=True)
            germline_array = airr.read_airr(self.germline_good, validate=True, debug=True)
            genotype_array = airr.read_airr(self.genotype_good, validate=True, debug=True)

            # Build keyed representation
            repertoire_data = {'Repertoire': {x['repertoire_id']: x for x in repertoire_array['Repertoire']}}
            germline_data = {'GermlineSet': {x['germline_set_id']: x for x in germline_array['GermlineSet']}}
            genotype_data = {'GenotypeSet': {x['receptor_genotype_set_id']: x for x in genotype_array['GenotypeSet']}}

            # combine together and write
            obj = {}
            obj['Repertoire'] = repertoire_data['Repertoire']
            obj['GermlineSet'] = germline_data['GermlineSet']
            obj['GenotypeSet'] = genotype_data['GenotypeSet']
            airr.write_airr(self.output_good, obj, validate=True, debug=True)

            # verify we can read it
            data = airr.read_airr(self.output_good, validate=True, debug=True)

            # is the data identical?
            del data['Info']
            if jsondiff.diff(obj, data) != {}:
                print('Output data does not match', file=sys.stderr)
                print(jsondiff.diff(obj, data), file=sys.stderr)
                self.assertTrue(False, 'write_airr_data(): Output data does not match')

        except Exception as inst:
            self.assertTrue(False, 'write_airr_data(): good data failed')
            print(type(inst))
            raise inst


if __name__ == '__main__':
    unittest.main()



"""
Reference library for AIRR schema for Ig/TCR rearrangements
"""
from __future__ import print_function
import sys
import csv
from airr.schema import RearrangementSchema, ValidationError


class RearrangementReader:
    """
    Iterator for reading Rearrangement objects in TSV format

    Attributes:
      fields (list): field names in the input Rearrangement file.
      external_fields (list): list of fields in the input file that are not
                              part of the Rearrangement definition.
    """
    @property
    def fields(self):
        """
        Get list of fields

        Returns:
          list : field names.
        """
        return self.dict_reader.fieldnames

    @property
    def external_fields(self):
        """
        Get list of field that are not in the Rearrangement schema

        Returns:
          list : field names.
        """
        return [f for f in self.dict_reader.fieldnames \
                if f not in self.schema.properties]

    def __init__(self, handle, base=1, validate=False, debug=False):
        """
        Initialization

        Arguments:
          handle (file): file handle of the open Rearrangement file.
          base (int): one of 0 or 1 specifying the coordinate schema in the input file.
                      If 1, then the file is assumed to contain 1-based closed intervals
                      that will be converted to python style 0-based half-open intervals
                      for known fields. If 0, then values will be unchanged.
          validate (bool): perform validation. If True then basic validation will be
                           performed will reading the data. A ValidationError exception
                           will be raised if an error is found.
          debug (bool): debug state. If True prints debug information.

        Returns:
          airr.io.RearrangementReader: reader object.
        """
        # arguments
        self.handle = handle
        self.base = base
        self.debug = debug
        self.validate = validate
        self.schema = RearrangementSchema

        # data reader, collect field names
        self.dict_reader = csv.DictReader(self.handle, dialect='excel-tab')

    def __iter__(self):
        """
        Iterator initializer

        Returns:
          airr.io.RearrangementReader
        """
        # Validate fields
        if (self.validate):
            self.schema.validate_header(self.dict_reader.fieldnames)

        return self

    def __next__(self):
        """
        Next method

        Returns:
          dict: parsed Rearrangement data.
        """
        try:
            row = next(self.dict_reader)
        except StopIteration:
            raise StopIteration

        for f in row:
            # row entry with no header
            if f is None:
                if self.validate:
                    raise ValidationError('row has extra data')
                else:
                    raise ValueError('row has extra data')

            # Convert types
            spec = self.schema.type(f)
            try:
                if spec == 'boolean':
                    row[f] = self.schema.to_bool(row[f], validate=self.validate)
                if spec == 'integer':
                    row[f] = self.schema.to_int(row[f], validate=self.validate)
                if spec == 'number':
                    row[f] = self.schema.to_float(row[f], validate=self.validate)
            except ValidationError as e:
                raise ValidationError('field %s has %s' %(f, e))

            # Adjust coordinates
            if f and f.endswith('_start') and self.base == 1:
                try:
                    row[f] = row[f] - 1
                except TypeError:
                    row[f] = None

        return row

    def close(self):
        """
        Closes the Rearrangement file
        """
        self.handle.close()

    def next(self):
        """
        Next method
        """
        return self.__next__()


class RearrangementWriter:
    """
    Writer class for Rearrangement objects in TSV format

    Attributes:
      fields (list): field names in the output Rearrangement file.
      external_fields (list): list of fields in the output file that are not
                              part of the Rearrangement definition.
    """
    @property
    def fields(self):
        """
        Get list of fields

        Returns:
          list : field names.
        """
        return self.dict_writer.fieldnames

    @property
    def external_fields(self):
        """
        Get list of field that are not in the Rearrangements schema

        Returns:
          list : field names.
        """
        return [f for f in self.dict_writer.fieldnames \
                if f not in self.schema.properties]

    def __init__(self, handle, fields=None, base=1, debug=False):
        """
        Initialization

        Arguments:
          handle (file): file handle of the open Rearrangements file.
          fields (list) : list of non-required fields to add. May include fields undefined by the schema.
          base (int): one of 0 or 1 specifying the coordinate schema in the output file.
                      Data provided to the write is assumed to be in python style 0-based
                      half-open intervals. If 1, then data will be converted to 1-based
                      closed intervals for known fields before writing. If 0, then values will be unchanged.
          debug (bool): debug state. If True prints debug information.

        Returns:
          airr.io.RearrangementWriter: writer object.
        """
        # arguments
        self.handle = handle
        self.base = base
        self.debug = debug
        self.schema = RearrangementSchema

        # order fields according to spec
        field_names = list(self.schema.required)
        if fields is not None:
            additional_fields = []
            for f in fields:
                if f in self.schema.required:
                    continue
                elif f in self.schema.optional:
                    field_names.append(f)
                else:
                    additional_fields.append(f)
            field_names.extend(additional_fields)

        # open writer and write header
        self.dict_writer = csv.DictWriter(self.handle, fieldnames=field_names, dialect='excel-tab',
                                          extrasaction='ignore', lineterminator='\n')
        self.dict_writer.writeheader()

    def close(self):
        """
        Closes the Rearrangement file
        """
        self.handle.close()

    def write(self, row):
        """
        Write a row to the Rearrangement file

        Arguments:
            row (dict): row to write.
        """
        # validate row
        if self.debug:
            for field in self.schema.required:
                if row.get(field, None) is None:
                    sys.stderr.write('Warning: Record is missing AIRR required field (' + field + ').\n')

        for f in row.keys():
            # Adjust coordinates
            if f.endswith('_start') and self.base == 1:
                try:
                    row[f] = self.schema.to_int(row[f]) + 1
                except TypeError:
                    row[f] = None

            # Convert types
            spec = self.schema.type(f)
            if spec == 'boolean':  row[f] = self.schema.from_bool(row[f])

        self.dict_writer.writerow(row)


# TODO: pandas validation need if we load with pandas directly
# def validate_df(df, airr_schema):
#     valid = True
#
#     # check required fields
#     missing_fields = set(airr_schema.required) - set(df.columns)
#     if len(missing_fields) > 0:
#         print('Warning: file is missing mandatory fields: {}'.format(', '.join(missing_fields)))
#         valid = False
#
#     if not valid:
#         raise ValueError('invalid AIRR data file')





"""
Unit tests for interface
"""
# System imports
import os
import time
import unittest

# Load imports
from airr.io import *

# Paths
test_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(test_path, 'data')


class TestRearrangementReader(unittest.TestCase):
    def setUp(self):
        print('-------> %s()' % self.id())

        # Test data
        self.data_good = os.path.join(data_path, 'good_rearrangement.tsv')
        self.data_bad = os.path.join(data_path, 'bad_rearrangement.tsv')
        self.data_extra = os.path.join(data_path, 'extra_rearrangement.tsv')

        # Start timer
        self.start = time.time()

    def tearDown(self):
        t = time.time() - self.start
        print('<- %.3f %s()' % (t, self.id()))

    # @unittest.skip('-> validate(): skipped\n')
    def test_validate(self):
        # Good data
        try:
            with open(self.data_good, 'r') as handle:
                reader = RearrangementReader(handle, validate=True)
                for r in reader:
                    pass
        except:
            self.assertTrue(False, 'validate(): good data failed')

        # Bad data
        try:
            with open(self.data_bad, 'r') as handle:
                reader = RearrangementReader(handle, validate=True)
                for r in reader:
                    pass
            self.assertFalse(True, 'validate(): bad data failed')
        except ValidationError:
            pass
        except Exception as inst:
            print(type(inst))
            raise inst

        # Extra data
        try:
            with open(self.data_extra, 'r') as handle:
                reader = RearrangementReader(handle, validate=False)
                for r in reader:
                    pass
            self.assertFalse(True, 'validate(): extra data failed')
        except ValueError:
            pass
        except Exception as inst:
            print(type(inst))
            raise inst


if __name__ == '__main__':
    unittest.main()



"""
AIRR community formats for adaptive immune receptor data.
"""
import sys
import versioneer

try:
    from setuptools import setup, find_packages
except ImportError:
    sys.exit('setuptools is required.')

with open('README.rst', 'r') as ip:
    long_description = ip.read()

# Parse requirements
with open('requirements.txt') as req:
        install_requires = req.read().splitlines()

# Setup
setup(name='airr',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      author='AIRR Community',
      author_email='',
      description='AIRR Community Data Representation Standard reference library for antibody and TCR sequencing data.',
      long_description=long_description,
      zip_safe=False,
      license='CC BY 4.0',
      url='http://docs.airr-community.org',
      keywords=['AIRR', 'bioinformatics', 'sequencing', 'immunoglobulin', 'antibody',
                'adaptive immunity', 'T cell', 'B cell', 'BCR', 'TCR'],
      install_requires=install_requires,
      packages=find_packages(),
      package_data={'airr': ['specs/*.yaml']},
      entry_points={'console_scripts': ['airr-tools=airr.tools:main']},
      classifiers=['Intended Audience :: Science/Research',
                   'Natural Language :: English',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python :: 3',
                   'Topic :: Scientific/Engineering :: Bio-Informatics'])



"""
Interface functions for file operations
"""
from __future__ import absolute_import

# System imports
import gzip
import json
import sys
import pandas as pd
import yaml
import yamlordereddictloader
from collections import OrderedDict
from itertools import chain
from io import open
from warnings import warn

if (sys.version_info > (3, 0)):
    from io import StringIO
else:
    # Python 2 code in this block
    from io import BytesIO as StringIO

# Load imports
from airr.io import RearrangementReader, RearrangementWriter
from airr.schema import Schema, RearrangementSchema, RepertoireSchema, AIRRSchema, DataFileSchema, ValidationError

#### Rearrangement ####

def read_rearrangement(filename, validate=False, debug=False):
    """
    Open an iterator to read an AIRR rearrangements file

    Arguments:
      file (str): path to the input file.
      validate (bool): whether to validate data as it is read, raising a ValidationError
                       exception in the event of an error.
      debug (bool): debug flag. If True print debugging information to standard error.

    Returns:
      airr.io.RearrangementReader: iterable reader class.
    """
    if filename.endswith(".gz"):
        handle = gzip.open(filename, 'r')
    else:
        handle = open(filename, 'r')
        
    return RearrangementReader(handle, validate=validate, debug=debug)


def create_rearrangement(filename, fields=None, debug=False):
    """
    Create an empty AIRR rearrangements file writer

    Arguments:
      filename (str): output file path.
      fields (list): additional non-required fields to add to the output.
      debug (bool): debug flag. If True print debugging information to standard error.

    Returns:
      airr.io.RearrangementWriter: open writer class.
    """
    return RearrangementWriter(open(filename, 'w+'), fields=fields, debug=debug)


def derive_rearrangement(out_filename, in_filename, fields=None, debug=False):
    """
    Create an empty AIRR rearrangements file with fields derived from an existing file

    Arguments:
      out_filename (str): output file path.
      in_filename (str): existing file to derive fields from.
      fields (list): additional non-required fields to add to the output.
      debug (bool): debug flag. If True print debugging information to standard error.

    Returns:
      airr.io.RearrangementWriter: open writer class.
    """
    reader = RearrangementReader(open(in_filename, 'r'))
    in_fields = list(reader.fields)
    if fields is not None:
        in_fields.extend([f for f in fields if f not in in_fields])

    return RearrangementWriter(open(out_filename, 'w+'), fields=in_fields, debug=debug)


def load_rearrangement(filename, validate=False, debug=False):
    """
    Load the contents of an AIRR rearrangements file into a data frame

    Arguments:
      filename (str): input file path.
      validate (bool): whether to validate data as it is read, raising a ValidationError
                       exception in the event of an error.
      debug (bool): debug flag. If True print debugging information to standard error.

    Returns:
      pandas.DataFrame: Rearrangement records as rows of a data frame.
    """
    # TODO: test pandas.DataFrame.read_csv with converters argument as an alterative
    schema = RearrangementSchema

    df = pd.read_csv(filename, sep='\t', header=0, index_col=None,
                     dtype=schema.pandas_types(), true_values=schema.true_values,
                     false_values=schema.false_values)
    # added to use RearrangementReader without modifying it:
    buffer = StringIO()  # create an empty buffer
    df.to_csv(buffer, sep='\t', index=False)  # fill buffer
    buffer.seek(0)  # set to the start of the stream

    reader = RearrangementReader(buffer, validate=validate, debug=debug)

    df = pd.DataFrame(list(reader))
    return df


def dump_rearrangement(dataframe, filename, debug=False):
    """
    Write the contents of a data frame to an AIRR rearrangements file

    Arguments:
      dataframe (pandas.DataFrame): data frame of rearrangement data.
      filename (str): output file path.
      debug (bool): debug flag. If True print debugging information to standard error.

    Returns:
      bool: True if the file is written without error.
    """
    # TODO: test pandas.DataFrame.to_csv with converters argument as an alterative
    # dataframe.to_csv(handle, sep='\t', header=True, index=False, encoding='utf-8')

    fields = dataframe.columns.tolist()
    with open(filename, 'w+') as handle:
        writer = RearrangementWriter(handle, fields=fields, debug=debug)
        for __, row in dataframe.iterrows():
            writer.write(row.to_dict())

    return True


def merge_rearrangement(out_filename, in_filenames, drop=False, debug=False):
    """
    Merge one or more AIRR rearrangements files

    Arguments:
      out_filename (str): output file path.
      in_filenames (list): list of input files to merge.
      drop (bool): drop flag. If True then drop fields that do not exist in all input
                   files, otherwise combine fields from all input files.
      debug (bool): debug flag. If True print debugging information to standard error.

    Returns:
      bool: True if files were successfully merged, otherwise False.
    """
    try:
        # gather fields from input files
        readers = (RearrangementReader(open(f, 'r'), debug=False) for f in in_filenames)
        field_list = [x.fields for x in readers]
        if drop:
            field_set = set.intersection(*map(set, field_list))
        else:
            field_set = set.union(*map(set, field_list))
        field_order = OrderedDict([(f, None) for f in chain(*field_list)])
        out_fields = [f for f in field_order if f in field_set]

        # write input files to output file sequentially
        readers = (RearrangementReader(open(f, 'r'), debug=debug) for f in in_filenames)
        with open(out_filename, 'w+') as handle:
            writer = RearrangementWriter(handle, fields=out_fields, debug=debug)
            for reader in readers:
                for r in reader:  writer.write(r)
                reader.close()
    except Exception as e:
        sys.stderr.write('Error occurred while merging AIRR rearrangement files: %s\n' % e)
        return False

    return True


def validate_rearrangement(filename, debug=False):
    """
    Validates an AIRR rearrangements file

    Arguments:
      filename (str): path of the file to validate.
      debug (bool): debug flag. If True print debugging information to standard error.

    Returns:
      bool: True if files passed validation, otherwise False.
    """
    valid = True
    if debug:
        sys.stderr.write('Validating: %s\n' % filename)

    # Open reader
    handle = open(filename, 'r')
    reader = RearrangementReader(handle, validate=True)

    # Validate header
    try:
        iter(reader)
    except ValidationError as e:
        valid = False
        if debug:
            sys.stderr.write('%s has validation error: %s\n' % (filename, e))

    # Validate each row
    i = 0
    while True:
        try:
            i = i + 1
            next(reader)
        except StopIteration:
            break
        except ValidationError as e:
            valid = False
            if debug:
                sys.stderr.write('%s at record %i has validation error: %s\n' % (filename, i, e))

    # Close
    handle.close()

    return valid

#### AIRR Data Model ####

def read_airr(filename, format=None, validate=False, model=True, debug=False):
    """
    Load an AIRR Data file

    Arguments:
      filename (str): path to the input file.
      format (str): input file format valid strings are "yaml" or "json". If set to None,
                    the file format will be automatically detected from the file extension.
      validate (bool): whether to validate data as it is read, raising a ValidationError
                       exception in the event of a validation failure.
      model (bool): If True only validate objects defined in the AIRR DataFile schema.
                  If False, attempt validation of all top-level objects.
                  Ignored if validate=False.
      debug (bool): debug flag. If True print debugging information to standard error.

    Returns:
      dict: dictionary of AIRR Data objects.
    """
    # Because the AIRR Data File is read in completely, we do not bother with a reader class.
    # Determine file type from extension and use appropriate loader
    ext = str.lower(filename.split('.')[-1]) if not format else format
    if ext in ('yaml', 'yml'):
        with open(filename, 'r', encoding='utf-8') as handle:
            data = yaml.load(handle, Loader=yamlordereddictloader.Loader)
    elif ext == 'json':
        with open(filename, 'r', encoding='utf-8') as handle:
            data = json.load(handle)
    else:
        if debug:  sys.stderr.write('Unknown file type: %s. Supported file extensions are "yaml", "yml" or "json"\n' % ext)
        raise TypeError('Unknown file type: %s. Supported file extensions are "yaml", "yml" or "json"\n' % ext)
        data = None

    # Validate if requested
    if validate:
        if debug:  sys.stderr.write('Validating: %s\n' % filename)
        try:
            valid = validate_airr(data, model=model, debug=debug)
        except ValidationError as e:
            if debug:  sys.stderr.write('%s failed validation\n' % filename)
            raise ValidationError(e)

    # We do not perform any additional processing
    return data


def validate_airr(data, model=True, debug=False):
    """
    Validates an AIRR Data file

    Arguments:
      data (dict): dictionary containing AIRR Data Model objects
      model (bool): If True only validate objects defined in the AIRR DataFile schema.
                  If False, attempt validation of all top-level objects
      debug (bool): debug flag. If True print debugging information to standard error.

    Returns:
      bool: True if files passed validation, otherwise False.
    """
    # Type check that input type is either dict or OrderedDict
    if not hasattr(data, 'items'):
        if debug:  sys.stderr.write('Data parameter is not a dictionary\n')
        raise TypeError('Data parameter is not a dictionary')

    # Loop through each AIRR object and validate
    valid = True
    for k, object in data.items():
        if k in ('Info', 'DataFile'):  continue
        if not object:  continue

        # Check for DataFile schema
        if model and k not in DataFileSchema.properties:
            if debug:  sys.stderr.write('Skipping non-DataFile object: %s\n' % k)
            continue

        # Get Schema
        schema = AIRRSchema.get(k, Schema(k))

        # Determine input type and set appropriate iterator
        if hasattr(object, 'items'):
            # Validate named array (dict)
            obj_iter = object.items()
            # Validate named array (dict) or a single object (dict)
            # obj_iter = object.items() if 'definition' not in object.keys() else [0, object]
        elif isinstance(object, list):
            # Validate array
            obj_iter = enumerate(object)
        else:
            # Unrecognized data structure
            valid = False
            if debug:  sys.stderr.write('%s is an unrecognized data structure: %s\n' % k)
            continue

        # Validate each record in array
        for i, record in obj_iter:
            try:
                schema.validate_object(record)
            except ValidationError as e:
                valid = False
                if debug:  sys.stderr.write('%s at array position %s with validation error: %s\n' % (k, i, e))

    if not valid:
        raise ValidationError('AIRR Data Model has validation failures')

    return valid


def write_airr(filename, data, format=None, info=None, validate=False, model=True, debug=False):
    """
    Write an AIRR Data file

    Arguments:
      filename (str): path to the output file.
      data (dict): dictionary of AIRR Data Model objects.
      format (str): output file format valid strings are "yaml" or "json". If set to None,
                    the file format will be automatically detected from the file extension.
      info (object): info object to write. Will write current AIRR Schema info if not specified.
      validate (bool): whether to validate data before it is written, raising a ValidationError
                       exception in the event of a validation failure.
      model (bool): If True only validate and write objects defined in the AIRR DataFile schema.
                  If False, attempt validation and write of all top-level objects
      debug (bool): debug flag. If True print debugging information to standard error.

    Returns:
      bool: True if the file is written without error.
    """
    # Type check that input type is either dict or OrderedDict
    if not hasattr(data, 'items'):
        if debug:  sys.stderr.write('Data parameter is not a dictionary\n')
        raise TypeError('Data parameter is not a dictionary')

    # Validate if requested
    if validate:
        if debug:  sys.stderr.write('Validating: %s\n' % filename)
        try:
            valid = validate_airr(data, model=model, debug=debug)
        except ValidationError as e:
            if debug:  sys.stderr.write(e)
            raise ValidationError(e)

    md = OrderedDict()
    if info is None:
        info = RearrangementSchema.info.copy()
        info['title'] = 'AIRR Data File'
        info['description'] = 'AIRR Data File written by AIRR Standards Python Library'
    md['Info'] = info

    # Loop through each object and add them to the output dict
    for k, obj in data.items():
        if k in ('Info', 'DataFile'):  continue
        if not obj:  continue
        if model and k not in DataFileSchema.properties:
            if debug:  sys.stderr.write('Skipping non-DataFile object: %s\n' % k)
            continue
        md[k] = obj

    # Determine file type from extension and use appropriate loader
    ext = str.lower(filename.split('.')[-1]) if not format else format
    if ext in ('yaml', 'yml'):
        with open(filename, 'w') as handle:
            yaml.dump(md, handle, default_flow_style=False)
    elif ext == 'json':
        with open(filename, 'w') as handle:
            json.dump(md, handle, sort_keys=False, indent=2)
    else:
        if debug:
            sys.stderr.write('Unknown file type: %s. Supported file extensions are "yaml", "yml" or "json"\n' % ext)
        raise TypeError('Unknown file type: %s. Supported file extensions are "yaml", "yml" or "json"\n' % ext)

    return True


#### Deprecated ####

def repertoire_template():
    """
    Return a blank repertoire object from the template. This object has the complete
    structure with all of the fields and all values set to None or empty string.

    Returns:
      object: empty repertoire object.

    .. deprecated:: 1.4
       Use :meth:`schema.Schema.template` instead.
    """
    # Deprecation
    warn('repertoire_template is deprecated and will be removed in a future release.\nUse schema.Schema.template instead.\n',
         DeprecationWarning, stacklevel=2)

    # Build template
    object = RepertoireSchema.template()

    return object


def load_repertoire(filename, validate=False, debug=False):
    """
    Load an AIRR repertoire metadata file

    Arguments:
      filename (str): path to the input file.
      validate (bool): whether to validate data as it is read, raising a ValidationError
                       exception in the event of an error.
      debug (bool): debug flag. If True print debugging information to standard error.

    Returns:
      dict: dictionary of AIRR Data objects.

    .. deprecated:: 1.4
       Use :func:`read_airr` instead.
    """
    # Deprecation
    warn('load_repertoire is deprecated and will be removed in a future release.\nUse read_airr instead.\n',
         DeprecationWarning, stacklevel=2)

   # use standard load function, we only validate Repertoire if requested
    md = read_airr(filename, validate=validate, debug=debug)

    if md.get('Repertoire') is None:
        if debug:
            sys.stderr.write('%s is missing "Repertoire" key\n' % (filename))
        raise KeyError('Repertoire object cannot be found in the file')

    # validate if requested
    if validate:
        valid = True
        reps = md['Repertoire']
        i = 0
        for r in reps:
            try:
                RepertoireSchema.validate_object(r)
            except ValidationError as e:
                valid = False
                if debug:
                    sys.stderr.write('%s has repertoire at array position %i with validation error: %s\n' % (filename, i, e))
            i = i + 1
        if not valid:
            raise ValidationError('Repertoire file %s has validation errors\n' % (filename))

    # we do not perform any additional processing
    return md


def validate_repertoire(filename, debug=False):
    """
    Validates an AIRR repertoire metadata file

    Arguments:
      filename (str): path of the file to validate.
      debug (bool): debug flag. If True print debugging information to standard error.

    Returns:
      bool: True if files passed validation, otherwise False.

    .. deprecated:: 1.4
       Use :func:`validate_airr` instead.
    """
    # Deprecation
    warn('validate_repertoire is deprecated and will be removed in a future release.\nUse validate_airr instead.\n',
         DeprecationWarning, stacklevel=2)

    valid = True
    if debug:
        sys.stderr.write('Validating: %s\n' % filename)

    # load with validate
    try:
        data = load_repertoire(filename, validate=True, debug=debug)
    except TypeError:
        valid = False
    except KeyError:
        valid = False
    except ValidationError as e:
        valid = False
        if debug:
            sys.stderr.write('%s has validation error: %s\n' % (filename, e))

    return valid

def write_repertoire(filename, repertoires, info=None, debug=False):
    """
    Write an AIRR repertoire metadata file

    Arguments:
      file (str): path to the output file.
      repertoires (list): array of repertoire objects.
      info (object): info object to write. Will write current AIRR Schema info if not specified.
      debug (bool): debug flag. If True print debugging information to standard error.

    Returns:
      bool: True if the file is written without error.

    .. deprecated:: 1.4
       Use :func:`write_airr` instead.
    """
    # Deprecation
    warn('write_repertoire is deprecated and will be removed in a future release.\nUse write_airr instead.\n',
         DeprecationWarning, stacklevel=2)

    if not isinstance(repertoires, list):
        if debug:
            sys.stderr.write('Repertoires parameter is not a list\n')
        raise TypeError('Repertoires parameter is not a list')

    md = OrderedDict()
    if info is None:
        info = RearrangementSchema.info.copy()
        info['title'] = 'Repertoire metadata'
        info['description'] = 'Repertoire metadata written by AIRR Standards Python Library'
    md['Info'] = info
    md['Repertoire'] = repertoires

    return write_airr(filename, md, info=info, debug=debug)



"""
AIRR Data Representation Schema
"""

# Imports
import sys
import yaml
import yamlordereddictloader
from collections import OrderedDict
from pkg_resources import resource_stream

with resource_stream(__name__, 'specs/airr-schema.yaml') as f:
    DEFAULT_SPEC = yaml.load(f, Loader=yamlordereddictloader.Loader)


class ValidationError(Exception):
    """
    Exception raised when validation errors are encountered.
    """
    pass


class Schema:
    """
    AIRR schema definitions

    Attributes:
      definition: name of the schema definition.
      info (collections.OrderedDict): schema info.
      properties (collections.OrderedDict): field definitions.
      required (list): list of mandatory fields.
      optional (list): list of non-required fields.
      false_values (list): accepted string values for False.
      true_values (list): accepted values for True.
    """
    # Boolean list for pandas
    true_values = ['True', 'true', 'TRUE', 'T', 't', '1']
    false_values = ['False', 'false', 'FALSE', 'F', 'f', '0']

    # Generate dicts for booleans
    _to_bool_map = {x: True for x in true_values + [1, True]}
    _to_bool_map.update({x: False for x in false_values + [0, False]})
    _from_bool_map = {k: 'T' if v else 'F' for k, v in _to_bool_map.items()}

    def __init__(self, definition):
        """
        Initialization

        Arguments:
          definition (string): the schema definition to load.

        Returns:
          airr.schema.Schema : schema object.
        """
        # Info is not a valid schema
        if definition == 'Info':
            raise KeyError('Info is an invalid schema definition name')

        # Load object definition
        if isinstance(definition, dict):       # on-the-fly definition of a nested object
            self.definition = definition
            spec = {'Info': []}
        else:
            spec = DEFAULT_SPEC

            try:
                self.definition = spec[definition]
            except KeyError:
                raise KeyError('Schema definition %s cannot be found in the specifications' % definition)
            except:
                raise

        try:
            self.info = spec['Info']
        except KeyError:
            raise KeyError('Info object cannot be found in the specifications')
        except:
            raise

        if self.definition.get('properties') is not None:
            self.properties = self.definition['properties']
            try:
                self.required = self.definition['required']
            except KeyError:
                self.required = []
            except:
                raise
        elif self.definition.get('allOf') is not None:
            self.properties = {}
            self.required = []
            for s in self.definition['allOf']:
                if s.get('$ref') is not None:
                    schema_name = s['$ref'].split('/')[-1]
                    # cannot use cache here
                    schema = Schema(schema_name)
                    # no nested allOf ...
                    self.properties.update(schema.properties)
                    self.required.extend(schema.required)
                elif s.get('properties') is not None:
                    self.properties.update(s.get('properties'))
                    if s.get('required') is not None:
                        self.required.extend(s.get('required'))
                else:
                    raise KeyError('Cannot find properties for schema definition %s' % definition)

        self.optional = [f for f in self.properties if f not in self.required]

    def spec(self, field):
        """
        Get the properties for a field

        Arguments:
          name (str): field name.

        Returns:
          collections.OrderedDict: definition for the field.
        """
        return self.properties.get(field, None)

    def type(self, field):
        """
        Get the type for a field

        Arguments:
          name (str): field name.

        Returns:
          str: the type definition for the field
        """
        field_spec = self.properties.get(field, None)
        field_type = field_spec.get('type', None) if field_spec else None
        return field_type

    def pandas_types(self):
        """
        Map of schema types to pandas types

        Returns:
          dict: mapping dictionary for pandas types
        """
        type_mapping = {}
        for property in self.properties:
            if self.type(property) == 'boolean':
                type_mapping[property] = bool
            elif self.type(property) == 'integer':
                type_mapping[property] = 'Int64'
            elif self.type(property) == 'number':
                type_mapping[property] = 'float64'
            elif self.type(property) == 'string':
                type_mapping[property] = str

        return type_mapping

    def to_bool(self, value, validate=False):
        """
        Convert a string to a boolean

        Arguments:
          value (str): logical value as a string.
          validate (bool): when True raise a ValidationError for an invalid value.
                           Otherwise, set invalid values to None.

        Returns:
          bool: conversion of the string to True or False.

        Raises:
          airr.ValidationError: raised if value is invalid when validate is set True.
        """
        if value == '' or value is None:
            return None

        bool_value = self._to_bool_map.get(value, None)
        if bool_value is None and validate:
            raise ValidationError('invalid bool %s' % value)
        else:
            return bool_value

    def from_bool(self, value, validate=False):
        """
        Converts a boolean to a string

        Arguments:
          value (bool): logical value.
          validate (bool): when True raise a ValidationError for an invalid value.
                           Otherwise, set invalid values to None.

        Returns:
          str: conversion of True or False or 'T' or 'F'.

        Raises:
          airr.ValidationError: raised if value is invalid when validate is set True.
        """
        if value == '' or value is None:
            return ''

        str_value = self._from_bool_map.get(value, None)
        if str_value is None and validate:
            raise ValidationError('invalid bool %s' % value)
        else:
            return str_value

    def to_int(self, value, validate=False):
        """
        Converts a string to an integer

        Arguments:
          value (str): integer value as a string.
          validate (bool): when True raise a ValidationError for an invalid value.
                           Otherwise, set invalid values to None.

        Returns:
          int: conversion of the string to an integer.

        Raises:
          airr.ValidationError: raised if value is invalid when validate is set True.
        """
        if value == '' or value is None:
            return None
        if isinstance(value, int):
            return value

        try:
            return int(value)
        except ValueError:
            if validate:
                raise ValidationError('invalid int %s' % value)
            else:
                return None

    def to_float(self, value, validate=False):
        """
        Converts a string to a float

        Arguments:
          value (str): float value as a string.
          validate (bool): when True raise a ValidationError for an invalid value.
                           Otherwise, set invalid values to None.

        Returns:
          float: conversion of the string to a float.

        Raises:
          airr.ValidationError: raised if value is invalid when validate is set True.
        """
        if value == '' or value is None:
            return None
        if isinstance(value, float):
            return value

        try:
            return float(value)
        except ValueError:
            if validate:
                raise ValidationError('invalid float %s' % value)
            else:
                return None

    def validate_header(self, header):
        """
        Validate header against the schema

        Arguments:
          header (list): list of header fields.

        Returns:
          bool: True if a ValidationError exception is not raised.

        Raises:
          airr.ValidationError: raised if header fails validation.
        """
        # Check for missing header
        if header is None:
            raise ValidationError('missing header')

        # Check required fields
        missing_fields = [f for f in self.required if f not in header]

        if missing_fields:
            raise ValidationError('missing required fields (%s)' % ', '.join(missing_fields))
        else:
            return True

    def validate_row(self, row):
        """
        Validate Rearrangements row data against schema

        Arguments:
          row (dict): dictionary containing a single record.

        Returns:
          bool: True if a ValidationError exception is not raised.

        Raises:
          airr.ValidationError: raised if row fails validation.
        """
        for f in row:
            # Empty strings are valid
            if row[f] == '' or row[f] is None:
                continue

            # Check types
            spec = self.type(f)
            try:
                if spec == 'boolean': self.to_bool(row[f], validate=True)
                if spec == 'integer': self.to_int(row[f], validate=True)
                if spec == 'number': self.to_float(row[f], validate=True)
            except ValidationError as e:
                raise ValidationError('field %s has %s' % (f, e))

        return True

    def validate_object(self, obj, missing=True, nonairr=True, context=None):
        """
        Validate Repertoire object data against schema

        Arguments:
          obj (dict): dictionary containing a single repertoire object.
          missing (bool): provides warnings for missing optional fields.
          nonairr (bool: provides warning for non-AIRR fields that cannot be validated.
          context (string): used by recursion to indicate place in object hierarchy

        Returns:
          bool: True if a ValidationError exception is not raised.

        Raises:
          airr.ValidationError: raised if object fails validation.
        """
        # object has to be a dictionary
        if not hasattr(obj, 'items'):
            if context is None:
                raise ValidationError('object is not a dictionary')
            else:
                raise ValidationError('field "%s" is not a dictionary object' % context)

        # first warn about non-AIRR fields
        if nonairr:
            for f in obj:
                if context is None: full_field = f
                else: full_field = context + '.' + f
                if self.properties.get(f) is None:
                    sys.stderr.write('Warning: Object has non-AIRR field that cannot be validated (' + full_field + ').\n')

        # now walk through schema and check types
        for f in self.properties:
            if context is None: full_field = f
            else: full_field = context + '.' + f
            spec = self.spec(f)
            xairr = spec.get('x-airr')

            # check if deprecated
            if xairr and xairr.get('deprecated'):
                continue

            # check if null and if key is missing
            is_missing_key = False
            is_null = False
            if obj.get(f) is None:
                is_null = True
                if obj.get(f, 'missing') == 'missing':
                    is_missing_key = True

            # check MiAIRR keys exist
            if xairr and xairr.get('miairr'):
                if is_missing_key:
                    raise ValidationError('MiAIRR field "%s" is missing' % full_field)

            # check if required field
            if f in self.required and is_missing_key:
                raise ValidationError('Required field "%s" is missing' % full_field)

            # check if identifier field
            if xairr and xairr.get('identifier'):
                if is_missing_key:
                    if xairr.get('nullable'):
                        sys.stderr.write(
                            'Warning: Nullable identifier field "%s" is missing.\n' % full_field)
                    else:
                        raise ValidationError('Not-nullable identifier field "%s" is missing' % full_field)

            # check nullable requirements
            if is_null:
                if not xairr:
                    # default is true
                    continue
                if xairr.get('nullable') or xairr.get('nullable', 'missing') == 'missing':
                    # nullable is allowed
                    continue
                else:
                    # nullable not allowed
                    raise ValidationError('Non-nullable field "%s" is null or missing' % full_field)

            # if get to here, field should exist with non null value

            # check types
            field_type = self.type(f)
            if field_type is None:
                # for referenced object, recursively call validate with object and schema
                if spec.get('$ref') is not None:
                    schema_name = spec['$ref'].split('/')[-1]
                    if AIRRSchema.get(schema_name):
                        schema = AIRRSchema[schema_name]
                    else:
                        schema = Schema(schema_name)
                    schema.validate_object(obj[f], missing, nonairr, full_field)
                else:
                    raise ValidationError('Internal error: field "%s" in schema not handled by validation. File a bug report.' % full_field)
            elif field_type == 'array':
                if not isinstance(obj[f], list):
                    raise ValidationError('field "%s" is not an array' % full_field)

                # for array, check each object in it
                for row in obj[f]:
                    if spec['items'].get('$ref') is not None:
                        schema_name = spec['items']['$ref'].split('/')[-1]
                        schema = Schema(schema_name)
                        schema.validate_object(row, missing, nonairr, full_field)
                    elif spec['items'].get('allOf') is not None:
                        for s in spec['items']['allOf']:
                            if s.get('$ref') is not None:
                                schema_name = s['$ref'].split('/')[-1]
                                if AIRRSchema.get(schema_name):
                                    schema = AIRRSchema[schema_name]
                                else:
                                    schema = Schema(schema_name)
                                schema.validate_object(row, missing, False, full_field)
                    elif spec['items'].get('enum') is not None:
                        if row not in spec['items']['enum']:
                            raise ValidationError('field "%s" has value "%s" not among possible enumeration values' % (full_field, row))
                    elif spec['items'].get('type') == 'string':
                        if not isinstance(row, str):
                            raise ValidationError('array field "%s" does not have string type: %s' % (full_field, row))
                    elif spec['items'].get('type') == 'boolean':
                        if not isinstance(row, bool):
                            raise ValidationError('array field "%s" does not have boolean type: %s' % (full_field, row))
                    elif spec['items'].get('type') == 'integer':
                        if not isinstance(row, int):
                            raise ValidationError('array field "%s" does not have integer type: %s' % (full_field, row))
                    elif spec['items'].get('type') == 'number':
                        if not isinstance(row, float) and not isinstance(row, int):
                            raise ValidationError('array field "%s" does not have number type: %s' % (full_field, row))
                    elif spec['items'].get('type') == 'object':
                        sub_schema = Schema({'properties': spec['items'].get('properties')})
                        sub_schema.validate_object(row, missing, nonairr, context)
                    else:
                        raise ValidationError('Internal error: array field "%s" in schema not handled by validation. File a bug report.' % full_field)
            elif field_type == 'object':
                # right now all arrays of objects use $ref
                raise ValidationError('Internal error: field "%s" in schema not handled by validation. File a bug report.' % full_field)
            else:
                # check basic types
                if field_type == 'string':
                    if not isinstance(obj[f], str):
                        raise ValidationError('Field "%s" does not have string type: %s' % (full_field, obj[f]))
                elif field_type == 'boolean':
                    if not isinstance(obj[f], bool):
                        raise ValidationError('Field "%s" does not have boolean type: %s' % (full_field, obj[f]))
                elif field_type == 'integer':
                    if not isinstance(obj[f], int):
                        raise ValidationError('Field "%s" does not have integer type: %s' % (full_field, obj[f]))
                elif field_type == 'number':
                    if not isinstance(obj[f], float) and not isinstance(obj[f], int):
                        raise ValidationError('Field "%s" does not have number type: %s' % (full_field, obj[f]))
                else:
                    raise ValidationError('Internal error: Field "%s" with type %s in schema not handled by validation. File a bug report.' % (full_field, field_type))

                # check basic types enums
                enums = spec.get('enum')

                if enums is not None:
                    field_value = obj[f]
                    if field_value not in enums:
                        raise ValidationError(
                            'field "%s" has value "%s" not among possible enumeration values %s' % (full_field, field_value, enums)
                        )

        return True

    def template(self):
        """
        Create an empty template object

        Returns:
          collections.OrderedDict: dictionary with all schema properties set as None or an empty list.
        """
        # Set defaults for each data type
        type_default = {'boolean': False, 'integer': 0, 'number': 0.0, 'string': '', 'array':[]}

        # Fetch schema template definition for a $ref string
        def _reference(ref):
            x = ref.split('/')[-1]
            schema = AIRRSchema.get(x, Schema(x))
            return(schema.template())

        # Get default value
        def _default(spec):
            if 'nullable' in spec['x-airr'] and not spec['x-airr']['nullable']:
                if 'enum' in spec:
                    return spec['enum'][0]
                else:
                    return type_default.get(spec['type'], None)
            else:
                return None

        # Populate empty object
        object = OrderedDict()
        for k, spec in self.properties.items():
            # Skip deprecated
            if 'x-airr' in spec and spec['x-airr'].get('deprecated', False):
                continue

            # Population values
            if '$ref' in spec:
                object[k] = _reference(spec['$ref'])
            elif spec['type'] == 'array':
                if '$ref' in spec['items']:
                    object[k] = [_reference(spec['items']['$ref'])]
                else:
                    object[k] = []
            elif 'x-airr' in spec:
                object[k] = _default(spec)
            else:
                object[k] = None

        return(object)


# Preloaded schema
AIRRSchema = {
    'Info': Schema('InfoObject'),
    'DataFile': Schema('DataFile'),
    'Alignment': Schema('Alignment'),
    'Rearrangement': Schema('Rearrangement'),
    'Repertoire': Schema('Repertoire'),
    'RepertoireGroup': Schema('RepertoireGroup'),
    'Ontology': Schema('Ontology'),
    'Study': Schema('Study'),
    'Subject': Schema('Subject'),
    'Diagnosis': Schema('Diagnosis'),
    'SampleProcessing': Schema('SampleProcessing'),
    'CellProcessing': Schema('CellProcessing'),
    'PCRTarget': Schema('PCRTarget'),
    'NucleicAcidProcessing': Schema('NucleicAcidProcessing'),
    'SequencingRun': Schema('SequencingRun'),
    'SequencingData': Schema('SequencingData'),
    'DataProcessing': Schema('DataProcessing'),
    'GermlineSet': Schema('GermlineSet'),
    'Acknowledgement': Schema('Acknowledgement'),
    'RearrangedSequence': Schema('RearrangedSequence'),
    'UnrearrangedSequence': Schema('UnrearrangedSequence'),
    'SequenceDelineationV': Schema('SequenceDelineationV'),
    'AlleleDescription': Schema('AlleleDescription'),
    'GenotypeSet': Schema('GenotypeSet'),
    'Genotype': Schema('Genotype'),
    'Cell': Schema('Cell'),
    'Clone': Schema('Clone')
}

InfoSchema = AIRRSchema['Info']
DataFileSchema = AIRRSchema['DataFile']
AlignmentSchema = AIRRSchema['Alignment']
RearrangementSchema = AIRRSchema['Rearrangement']
RepertoireSchema = AIRRSchema['Repertoire']
GermlineSetSchema = AIRRSchema['GermlineSet']
GenotypeSetSchema = AIRRSchema['GenotypeSet']
